from sclblonnx import add_output, add_input, add_node, node, empty_graph, add_constant, run, merge, split, display, \
    join, concat
import numpy as np
"""
Some rudimentary tests of the functions in merge.py; should be extended.

For example usage see example_merge.py in /examples.
"""


def test_merge():
    """
    Functional test of merge().
    """
    # Subgraph 1
    sg1 = empty_graph("Graph 1")
    n1 = node('Add', inputs=['x1', 'x2'], outputs=['sum'])
    sg1 = add_node(sg1, n1)
    sg1 = add_input(sg1, 'x1', "FLOAT", [1])
    sg1 = add_input(sg1, 'x2', "FLOAT", [1])
    sg1 = add_output(sg1, 'sum', "FLOAT", [1])

    # Subgraph 2
    sg2 = empty_graph("Graph 2")
    sg2 = add_constant(sg2, "const", np.array([7]), "FLOAT")
    n2 = node("Equal", inputs=['sum', 'const'], outputs=['equal'])
    sg2 = add_node(sg2, n2)

    sg2 = add_input(sg2, 'sum', "FLOAT", [1])
    sg2 = add_output(sg2, 'equal', "BOOL", [1])

    g = merge(sg1, sg2, outputs=["sum"], inputs=["sum"])

    data = {"x1": np.array([2]).astype(np.float32), "x2": np.array([5]).astype(np.float32)}
    result = run(g, inputs=data, outputs=["equal"])
    assert result[0], "Sum of 2 and 5 should be equal to constant 7. Merged failed."


def test_join():
    """
    Functional test for join:
    """
    g1 = empty_graph("G1")
    n1 = node('Add', inputs=['x1_1', 'x1_2'], outputs=['sum_1'], name="n1")
    g1 = add_input(g1, 'x1_1', "FLOAT", [1])
    g1 = add_input(g1, 'x1_2', "FLOAT", [1])
    g1 = add_output(g1, 'sum_1', "FLOAT", [1])
    g1 = add_node(g1, n1)

    g2 = empty_graph("G2")
    n2 = node('Add', inputs=['x2_1', 'x2_2'], outputs=['sum_2'], name="n2")
    g2 = add_input(g2, 'x2_1', "FLOAT", [1])
    g2 = add_input(g2, 'x2_2', "FLOAT", [1])
    g2 = add_output(g2, 'sum_2', "FLOAT", [1])
    g2 = add_node(g2, n2)

    g3 = empty_graph("G3")
    n3 = node('Add', inputs=['x3_1', 'x3_2'], outputs=['sum_3'], name="n3")
    g3 = add_input(g3, 'x3_1', "FLOAT", [1])
    g3 = add_input(g3, 'x3_2', "FLOAT", [1])
    g3 = add_output(g3, 'sum_3', "FLOAT", [1])
    g3 = add_node(g3, n3)

    g = join(g1, g2, g3, [("sum_1", "x3_1")], [("sum_2", "x3_2")])

    data = {
        "x1_1": np.array([1]).astype(np.float32),
        "x1_2": np.array([2]).astype(np.float32),
        "x2_1": np.array([3]).astype(np.float32),
        "x2_2": np.array([4]).astype(np.float32),
    }
    result = run(g, inputs=data, outputs=["sum_3"])
    assert result[0], "Sum of 1,2, 3, and 4 should be equal to constant 10. Join failed."


def test_split():
    """
    Functional test for split
    """
    g1 = empty_graph("G1")
    n1 = node('Add', inputs=['x1_1', 'x1_2'], outputs=['sum_1'], name="n1")
    g1 = add_input(g1, 'x1_1', "FLOAT", [1])
    g1 = add_input(g1, 'x1_2', "FLOAT", [1])
    g1 = add_output(g1, 'sum_1', "FLOAT", [1])
    g1 = add_node(g1, n1)

    g2 = empty_graph("G2")
    n2 = node('Add', inputs=['x2_1', 'x2_2'], outputs=['sum_2'], name="n2")
    g2 = add_input(g2, 'x2_1', "FLOAT", [1])
    g2 = add_input(g2, 'x2_2', "FLOAT", [1])
    g2 = add_output(g2, 'sum_2', "FLOAT", [1])
    g2 = add_node(g2, n2)

    g3 = empty_graph("G3")
    n3 = node('Add', inputs=['x3_1', 'x3_2'], outputs=['sum_3'], name="n3")
    g3 = add_input(g3, 'x3_1', "FLOAT", [1])
    g3 = add_input(g3, 'x3_2', "FLOAT", [1])
    g3 = add_output(g3, 'sum_3', "FLOAT", [1])
    g3 = add_node(g3, n3)

    g = split(g1, g2, g3, cg1_match=[("sum_1", "x2_2")], cg2_match=[("sum_1", "x3_1")])

    data = {
        "x1_1": np.array([1]).astype(np.float32),
        "x1_2": np.array([2]).astype(np.float32),
        "x2_1": np.array([3]).astype(np.float32),
        "x3_2": np.array([4]).astype(np.float32),
    }
    result = run(g, inputs=data, outputs=["sum_2", "sum_3"])
    assert result[0], "Sum of 1,2, and 3 should be equal to constant 6. Split failed."


def test_concat():
    """
    Functional test for concat
    """
    g1 = empty_graph("G1")
    n1 = node('Add', inputs=['x1_1', 'x1_2'], outputs=['sum_1'], name="node_name")
    g1 = add_input(g1, 'x1_1', "FLOAT", [1])
    g1 = add_input(g1, 'x1_2', "FLOAT", [1])
    g1 = add_output(g1, 'sum_1', "FLOAT", [1])
    g1 = add_node(g1, n1)

    g2 = empty_graph("G2")
    n2 = node('Add', inputs=['x2_1', 'x2_2'], outputs=['sum_2'], name="node_name")
    g2 = add_input(g2, 'x2_2', "FLOAT", [1])
    g2 = add_output(g2, 'sum_2', "FLOAT", [1])
    g2 = add_node(g2, n2)

    g = concat(g1, g2, False, True, edge_match=[("x1_2", "x2_1")])

    data = {
        "x1_1": np.array([2]).astype(np.float32),
        "x1_2": np.array([5]).astype(np.float32),
        "x2_2": np.array([5]).astype(np.float32)}
    result = run(g, inputs=data, outputs=["sum_1", "sum_2"])
    assert result[0], "Sum of 2 and 5 should be equal to constant 7. Concat failed."


# Run tests, all passes:
test_merge()
test_join()
test_split()
test_concat()










from sclblonnx import constant, empty_graph, node, add_node, add_constant, add_output, run, display, clean
import numpy as np


def test_constant():
    c = constant("", np.array([1,2,]), "FLOAT")
    assert not c, "Constant creation should have failed without a name."
    c = constant("constant", np.array([1,2,]), "NONE")
    assert not c, "Constant creation should have failed without a valid data type."
    c = constant("constant", np.array([1,2,]), "FLOAT")
    check = getattr(c, "output", False)
    assert check[0] == "constant", "Constant creation should have worked."


def test_add_constant():

    # Simple add graph
    g = empty_graph()
    n1 = node('Add', inputs=['x1', 'x2'], outputs=['sum'])
    g = add_node(g, n1)

    # Add input and constant
    g = add_constant(g, 'x1', np.array([1]), "INT64")
    g = add_constant(g, 'x2', np.array([5]), "INT64")

    # Output:
    g = add_output(g, 'sum', "INT64", [1])

    # This works, but seems to fail for other data types...
    result = run(g, inputs={}, outputs=["sum"])
    assert result[0] == 6, "Add constant failed."
    # todo(McK): Does not work for INT16 / INT8, check?


import numpy as np
from onnx import helper as xhelp
from onnx import onnx_ml_pb2 as xpb2

from sclblonnx.utils import _data_type, _print
from sclblonnx.node import add_node


# constant creates a constant node.
def constant(name: str,
             value: np.array,
             data_type: str,
             **kwargs):
    """ Create a constant node

    Args:
        name: Name of the (output value of the) constant node to determine the graph topology
        value: Values of the node (as a np.array)
        data_type: Data type of the node
        **kwargs

    Returns:
        A constant node.
    """
    if not name:
        _print("Unable to create unnamed constant.")
        return False

    dtype = _data_type(data_type)
    if not dtype:
        return False

    try:
        constant_node = xhelp.make_node('Constant', inputs=[], outputs=[name], name=name + "-constant",
                                        value=xhelp.make_tensor(name=name + "-values", data_type=dtype,
                                                                dims=value.shape, vals=value.flatten()), **kwargs)
    except Exception as e:
        _print("Unable to create the constant node: " + str(e))
        return False

    return constant_node


# add_constant adds a constant node to a graph
def add_constant(
        graph: xpb2.GraphProto,
        name: str,
        value: np.array,
        data_type: str,
        **kwargs):
    """ Create and add a constant node to an existing graph.

    Note: use add_node() if you want to add an existing constant node to an existing graph

    Args:
        graph: A graph, onnx.onnx_ml_pb2.GraphProto.
        name: Name of the (output value of the) constant node to determine the graph topology
        value: Values of the node (as a np.array)
        data_type: Data type of the node

    Returns:
        The extended graph.
    """
    if type(graph) is not xpb2.GraphProto:
        print("graph is not a valid ONNX graph.")
        return False

    dtype = _data_type(data_type)
    if not dtype:
        return False

    try:
        constant_node = xhelp.make_node('Constant', inputs=[], outputs=[name], name=name + "-constant",
                                        value=xhelp.make_tensor(name=name + "-values", data_type=dtype,
                                                                dims=value.shape, vals=value.flatten()), **kwargs)
    except Exception as e:
        _print("Unable to create the constant node: " + str(e))
        return False

    try:
        graph = add_node(graph, constant_node, **kwargs)
    except Exception as e:
        _print("Unable to add the constant node to the graph: " + str(e))
        return False

    if not graph:
        _print("Unable to add constant node to graph.")
        return False
    return graph




from onnx import onnx_ml_pb2 as xpb2
from sclblonnx import node, empty_graph, add_node, delete_node


def test_node():
    n1 = node('Add', inputs=['x1', 'x2'], outputs=['sum'])
    assert type(n1) is xpb2.NodeProto, "Error creating node."
    n2 = node('Add', inputs=['x1', 'x2'], outputs=['sum'], name="node_name")
    name = getattr(n2, "name", False)
    assert name == "node_name", "Node name should be node_name."


def test_add_node():
    g = empty_graph()
    n = node('Add', inputs=['x1', 'x2'], outputs=['sum'])
    g = add_node(g, n)
    assert len(g.node) == 1, "Node not properly added."


def test_add_nodes():
    g = empty_graph()
    for i in range(10):
        n = node('Add', inputs=['x1', 'x2'], outputs=['sum'])
        g = add_node(g, n)
    assert len(g.node) == 10, "Nodes not properly added."
    check = add_node(False, True)
    assert not check, "Incorrectly able to add to non-graph."
    check = add_node(g, False)
    assert not check, "Incorrectly able to add a non-node."


def test_delete_node():
    g = empty_graph()
    n = node('Add', inputs=['x1', 'x2'], outputs=['sum'], name="node_name")
    g = add_node(g, n)
    assert len(g.node) == 1, "Node not properly added."
    g = delete_node(g, "node_name")
    assert len(g.node) == 0, "Node not properly deleted."



from onnx import helper as xhelp
from onnx import onnx_ml_pb2 as xpb2

from sclblonnx.utils import _parse_element, _value, _data_type, _print


# list_outputs list all outputs in a graph
def list_outputs(graph: xpb2.GraphProto):
    """ Tries to list the outputs of a given graph.

    Args:
        graph the ONNX graph
    """
    if type(graph) is not xpb2.GraphProto:
        _print("graph is not a valid ONNX graph.")
        return False

    i = 1
    for elem in graph.output:
        name, dtype, shape = _parse_element(elem)
        print("Output {}: Name: '{}', Type: {}, Dimension: {}".format(i, name, dtype, shape))
        i += 1

    if i == 1:
        print("No outputs found.")

    return True


# add_output adds an output to a graph
def add_output(
        graph: xpb2.GraphProto,
        name: str,
        data_type: str,
        dimensions: [],
        **kwargs):
    """ Add an output to a graph

    Args:
        graph: A graph, onnx.onnx_ml_pb2.GraphProto.
        name: String, the name of the input as used to determine the graph topology.
        data_type: String, the data type of the input. Run list_data_types() for an overview.
        dimensions: List[] specifying the dimensions of the input.
        **kwargs

    Returns:
        The extended graph.

    """
    if type(graph) is not xpb2.GraphProto:
        _print("graph is not a valid ONNX graph.")
        return False

    dtype = _data_type(data_type)
    if not dtype:
        return False

    try:
        graph.output.append(xhelp.make_tensor_value_info(name, dtype, dimensions, **kwargs), **kwargs)
    except Exception as e:
        _print("Unable to add the input: " + str(e))
        return False
    return graph


# rename_output renames an existing output
def rename_output(graph, current_name, new_name):
    """ Rename an output to a graph

    Args:
        graph: A graph, onnx.onnx_ml_pb2.GraphProto.
        current_name: String, the current output name.
        new_name: String, the name desired output name.

    Returns:
        The changed graph.
    """
    if type(graph) is not xpb2.GraphProto:
        _print("graph is not a valid ONNX graph.")
        return False

    found = False
    for output in graph.output:
        if output.name == current_name:
            output.name = new_name
            found = True
    if not found:
        _print("Unable to found the output by name.")
        return False

    for node in graph.node:
        for index, name in enumerate(node.output):
            if name == current_name:
                node.output[index] = new_name

    return graph


# replace_output replaces an existing output
def replace_output(
        graph: xpb2.GraphProto,
        name: str,
        data_type: str,
        dimensions: [],
        **kwargs):
    """ Changes an existing output of a graph

    Args:
        graph: A graph, onnx.onnx_ml_pb2.GraphProto.
        name: String, the name of the output as used to determine the graph topology.
        data_type: String, the data type of the output. Run list_data_types() for an overview.
        dimensions: List[] specifying the dimensions of the input.
        **kwargs

    Returns:
        The extended graph.

    """
    if type(graph) is not xpb2.GraphProto:
        _print("graph is not a valid ONNX graph.")
        return graph

    # Remove the named output
    found = False
    try:
        for elem in graph.output:
            if elem.name == name:
                graph.output.remove(elem)
                found = True
    except Exception as e:
        _print("Unable to iterate the outputs. " + str(e))
        return False
    if not found:
        _print("Unable to find the output by name.")

    # Create the new value
    try:
        val = _value(name, data_type, dimensions, **kwargs)
    except Exception as e:
        _print("Unable to create value. " + str(e))
        return False

    # Add the value to the output
    try:
        graph.output.append(val, **kwargs)
    except Exception as e:
        _print("Unable to add the output: " + str(e))
        return False

    return graph


# delete_output deletes an existing output
def delete_output(
            graph: xpb2.GraphProto,
            name: str):
    """ Removes an existing output of a graph by name

    Args:
        graph: A graph, onnx.onnx_ml_pb2.GraphProto.
        name: String, the name of the output as used to determine the graph topology.


    Returns:
        The extended graph.

    """
    if type(graph) is not xpb2.GraphProto:
        _print("graph is not a valid ONNX graph.")
        return graph

    # Remove the named output
    found = False
    try:
        for elem in graph.output:
            if elem.name == name:
                graph.output.remove(elem)
                found = True
    except Exception as e:
        _print("Unable to iterate the outputs. " + str(e))
        return False
    if not found:
        _print("Unable to find the output by name.")
        return False

    return graph



from sclblonnx import empty_graph, list_inputs, add_input, rename_input, replace_input, delete_input


def test_list_inputs():
    g = empty_graph()
    assert list_inputs(g), "No inputs listed."
    g = add_input(g, "test", "FLOAT", [0])
    list_inputs(g)
    assert not list_inputs(False), "List inputs should be false."


def test_add_input():
    g = empty_graph()
    g = add_input(g, "test", "FLOAT", [0])
    name = getattr(g.input[0], "name", False)  # get the first input name:
    assert name == "test", "'test' should be in list of inputs."


def test_rename_input():
    g = empty_graph()
    g = add_input(g, "test", "FLOAT", [0])
    g = rename_input(g, "test", "new_name")
    name = getattr(g.input[0], "name", False)  # get the first input name:
    assert name == "new_name", "New name should be in list of inputs."


def test_replace_input():
    g = empty_graph()
    g = add_input(g, "test", "FLOAT", [0])
    g = replace_input(g, "test", "FLOAT", [10,10])

    type = getattr(g.input[0], "type", False)  # get the input type
    ttype = getattr(type, "tensor_type", False)
    shape = getattr(ttype, "shape", False)
    dim = getattr(shape, "dim", False)
    dim_val = getattr(dim[0], "dim_value", False)

    assert dim_val == 10, "New dimension should be 10"


def test_delete_input():
    g = empty_graph()
    g = add_input(g, "test", "FLOAT", [0])
    g = delete_input(g, "test")
    assert len(g.input) == 0, "There should not be any inputs after delete."


from sclblonnx import empty_graph, list_outputs, add_output, rename_output, replace_output, delete_output


def test_list_outputs():
    g = empty_graph()
    assert list_outputs(g), "No outputs listed."
    g = add_output(g, "test", "FLOAT", [0])
    list_outputs(g)
    assert not list_outputs(False), "List outputs should be false."


def test_add_output():
    g = empty_graph()
    g = add_output(g, "test", "FLOAT", [0])
    name = getattr(g.output[0], "name", False)  # get the first output name:
    assert name == "test", "'test' should be in list of outputs."


def test_rename_output():
    g = empty_graph()
    g = add_output(g, "test", "FLOAT", [0])
    g = rename_output(g, "test", "new_name")
    name = getattr(g.output[0], "name", False)  # get the first output name:
    assert name == "new_name", "New name should be in list of outputs."


def test_replace_output():
    g = empty_graph()
    g = add_output(g, "test", "FLOAT", [0])
    g = replace_output(g, "test", "FLOAT", [10,10])

    type = getattr(g.output[0], "type", False)  # get the output type
    ttype = getattr(type, "tensor_type", False)
    shape = getattr(ttype, "shape", False)
    dim = getattr(shape, "dim", False)
    dim_val = getattr(dim[0], "dim_value", False)

    assert dim_val == 10, "New dimension should be 10"


def test_delete_output():
    g = empty_graph()
    g = add_output(g, "test", "FLOAT", [0])
    g = delete_output(g, "test")
    assert len(g.output) == 0, "There should not be any outputs after delete."


from sclblonnx import empty_graph, add_output, add_input
from sclblonnx.utils import _parse_element, _value, _input_details, _output_details, _print, _load_version_info, \
    _data_type, _data_string
from sclblonnx._globals import ONNX_VERSION_INFO

def test__parse_element():
    g = empty_graph()
    dims = [4, 3, 7]
    g = add_output(g, 'sum', "FLOAT", dims)
    for elem in g.output:
        _, elem_type, _ = _parse_element(elem)
    assert elem_type == "FLOAT", "Element not properly parsed."


def test__value():
    v = _value("test_value", "FLOAT", [1,2,3])
    v_name = getattr(v, "name", False)
    assert v_name ==  "test_value", "Wrong value created."


def test__input_details():
    g = empty_graph()
    g = add_input(g, 'sum', "FLOAT", [4, 3, 7])
    in_det = _input_details(g)
    assert in_det['sum']['data_type'] == "FLOAT", "Input details not correct."


def test__output_details():
    g = empty_graph()
    g = add_output(g, 'sum', "FLOAT", [4, 3, 7])
    out_det = _output_details(g)
    assert out_det['sum']['data_type'] == "FLOAT", "Output details not correct."


def test__print():
    print("\n")
    _print("Red warning.")
    _print("Normal feedback", "MSG")
    _print("Green literal", "LIT")
    pass


def test__load_version_info():
    assert not ONNX_VERSION_INFO, "Should not be loaded."
    _load_version_info()
    assert not ONNX_VERSION_INFO, "Should be loaded."


def test__data_type():
    assert _data_type("FLOAT") == 1, "Float should be 1."
    assert not _data_type("BLA"), "Bla should not be a data type."


def test__data_string():
    assert _data_string(1) == "FLOAT", "Float should be 1."
    assert not _data_string(99), "99 should not be a data string."


from sclblonnx import empty_graph, node, add_node, add_input, add_output, check, clean
from onnx import onnx_ml_pb2 as xpb2


def test_clean():
    g = empty_graph()
    n1 = node('Add', inputs=['x1', 'x2'], outputs=['sum'])
    g = add_node(g, n1)
    g = add_input(g, 'x1', "FLOAT", [1])
    g = add_input(g, 'x2', "FLOAT", [1])
    g = add_output(g, 'sum', "FLOAT", [1])
    g = clean(g)
    assert type(g) == xpb2.GraphProto, "Clean failed."


def test_check():

    # Invalid, no input/output:
    g = empty_graph()
    n1 = node('Add', inputs=['x1', 'x2'], outputs=['sum'])
    g = add_node(g, n1)
    assert not check(g), "Graph is not complete."

    # Valid graph
    g = add_input(g, 'x1', "FLOAT", [1])
    g = add_input(g, 'x2', "FLOAT", [1])
    g = add_output(g, 'sum', "FLOAT", [1])
    assert check(g), "Graph should pass checks."

    # Invalid: None operator:
    g = empty_graph()
    n1 = node('None', inputs=['x1', 'x2'], outputs=['sum'])
    g = add_node(g, n1)
    g = add_input(g, 'x1', "FLOAT", [1])
    g = add_input(g, 'x2', "FLOAT", [1])
    g = add_output(g, 'sum', "FLOAT", [1])
    assert not check(g), "Graph should not pass checks."

    # Invalid: Dynamic size input
    g = empty_graph()
    n1 = node('Add', inputs=['x1', 'x2'], outputs=['sum'])
    g = add_node(g, n1)
    g = add_input(g, 'x1', "FLOAT", [])
    g = add_input(g, 'x2', "FLOAT", [1])
    g = add_output(g, 'sum', "FLOAT", [1])
    assert not check(g), "Graph should not pass checks."

    check(g, _sclbl_check=False, _onnx_check=False)
    check(g, _onnx_check=False)  # Operator check.


# Global variables for the sclblonnx package.
import os

# Dictionary containing details to check support
VERSION_INFO_LOCATION: str = os.path.dirname(os.path.realpath(__file__)) + "/supported_onnx.json"
ONNX_VERSION_INFO: dict = {}

# Node counter:
NODE_COUNT = 1

# Optimizer passes:
OPTIMIZER_PASSES = ['eliminate_deadend',
                    'eliminate_duplicate_initializer',
                    'eliminate_identity',
                    'eliminate_if_with_const_cond',
                    'eliminate_nop_cast',
                    'eliminate_nop_dropout',
                    'eliminate_nop_flatten',
                    'eliminate_nop_monotone_argmax',
                    'eliminate_nop_pad',
                    'eliminate_nop_transpose',
                    'eliminate_unused_initializer',
                    'extract_constant_to_initializer',
                    'fuse_add_bias_into_conv',
                    'fuse_bn_into_conv',
                    'fuse_consecutive_concats',
                    'fuse_consecutive_log_softmax',
                    'fuse_consecutive_reduce_unsqueeze',
                    'fuse_consecutive_squeezes',
                    'fuse_consecutive_transposes',
                    'fuse_matmul_add_bias_into_gemm',
                    'fuse_pad_into_conv',
                    'fuse_transpose_into_gemm',
                    'lift_lexical_references']

# Data types, see https://deeplearning4j.org/api/latest/onnx/Onnx.TensorProto.DataType.html
DATA_TYPES = {
    "FLOAT": 1,
    "UINT8": 2,
    "INT8": 3,
    "UINT16": 4,
    "INT16": 5,
    "INT32": 6,
    "INT64": 7,
    # "STRING" : 8,
    "BOOL": 9,
    "FLOAT16": 10,
    "DOUBLE": 11,
    "UINT32": 12,
    "UINT64": 13,
    "COMPLEX64": 14,
    "COMPLEX128": 15
}



import copy
from onnx import onnx_ml_pb2 as xpb2
from sclblonnx import check, delete_output, delete_input
from sclblonnx.utils import _print
"""
merge.py contains a number of utilities to merge / combine existing graphs. The functions merge(), join(), and split()
provide easy to use wrappers around the actual workhorse concat(). The concat function is versatile; please
see the example_merge.py script in /examples to 
"""


def merge(
        sg1: xpb2.GraphProto,
        sg2: xpb2.GraphProto,
        outputs: [] = None,
        inputs: [] = None,
        io_match: [] = None,
        complete: bool = True,
        _verbose: bool = True,
        **kwargs):
    """
    merge merges two graphs.

    Given subgraph sg1 and subgraph sg2 merge attempts to link the identified outputs of sg1 to the
    inputs of sg2 resulting in a graph in which sg1 is the parent of sg2.

    Merge expects two complete graphs (i.e., it expects sg1 and sg2 to pass check(). If you would like more
    flexible merge options or partial merge please see the concat function (merge is merely a constrained wrapper
    around concat).

    Note: The args inputs and outputs are present for legacy reasons, we recommend using io_match directly.

    Args:
        sg1: Subgraph 1, the parent.
        sg2: Subgraph 2, the child.
        outputs: (Optional) A list of strings containing the names of the outputs of sg1 that are matched to inputs (in order of the desired match).
        inputs: (Optional) A list of strings containing the names of the inputs of sg2 to which the outputs of sg1 are matched.
        io_match: (Optional) A list of names pairs [("out1","in1"), ("out2","in2"),...]. This is an alternative for the inputs/outputs arguments.
        complete: (Optional) Boolean indicating whether the resulting graph should be complete (i.e., should pass check). Default True.
        _verbose: (Optional) Boolean indicating whether or not verbose user feedback should be provided. Default True.
    Returns:
        The merged graph g, or False (with a printed error message) if something is wrong.
    """
    # immutable defaults:
    if inputs is None:
        inputs = []
    if outputs is None:
        outputs = []
    if io_match is None:
        io_match = []

    # prevent changes to original
    sg1 = copy.deepcopy(sg1)
    sg2 = copy.deepcopy(sg2)

    # Check the inputs:
    if type(sg1) is not xpb2.GraphProto:
        _print("Graph sg1 is not an ONNX graph.")
        return False
    if type(sg2) is not xpb2.GraphProto:
        _print("Graph sg2 is not an ONNX graph.")
        return False
    if len(outputs) != len(inputs):
        _print("The number of outputs and inputs do not match.")
        return False
    if len(inputs) > 0 and len(io_match) > 0:
        _print("Please use either the inputs/outputs arguments OR the io_match argument (not both).")
        return False

    # Construct IO pairs
    if len(inputs) > 0:
        _print("Constructing the io_match list from your input and output.", "MSG", (not _verbose))
        io_match = []
        for idx, val in enumerate(outputs):
            io_match.append((val, inputs[idx]))

    # Use concat to do the merge
    g = concat(sg1, sg2, io_match=io_match, complete=complete, **kwargs)
    if not g:
        _print("Graph merge failed. Please checkout concat for additional options.", "MSG", (not _verbose))

    return g


def join(
        pg1: xpb2.GraphProto,
        pg2: xpb2.GraphProto,
        cg: xpb2.GraphProto,
        pg1_match: [] = None,
        pg2_match: [] = None,
        complete: bool = True,
        _verbose: bool = True,
        **kwargs):
    """
    join takes two parent graphs (pg1 & pg2) and merges them with a child graph cg.

    Join matches the outputs of pg1 to the inputs of cg specified in pg1_match, and similarly for pg2 and pg2_match.
    Desired matches are specified in pairs: [("out1","in1"), ("out2","in2"),...].

    Join by default assumes the resulting joined graph to be complete. Join is merely a wrapper around concat (used
    twice). For more flexible combinations of graphs please see concat().

    Note: ONNX concat operations might give unexpected results if names of elements collide, please use postfix_names()
    to prevent this (and always critically inspect the resulting graph).

    Args:
        pg1: Parent graph 1.
        pg2: Parent graph 2.
        cg: Child graph, the graph that will join together pg1 and pg2.
        pg1_match: (Optional) List of pairs matching outputs of pg1 to inputs of cg. Default [].
        pg2_match: (Optional) List of pairs matching outputs of pg2 to inputs of cg. Default [].
        complete: (Optional) Boolean indicating whether the resulting graph should be complete (i.e., should pass check). Default True.
        _verbose: (Optional) Boolean indicating whether or not verbose user feedback should be provided. Default True.
    Returns:
        The joined graph g (of False is something fails along the way).
    """
    # immutable defaults:
    if pg1_match is None:
        pg1_match = []
    if pg2_match is None:
        pg2_match = []

    # prevent changes to original
    pg1 = copy.deepcopy(pg1)
    pg2 = copy.deepcopy(pg2)
    cg = copy.deepcopy(cg)

    if type(pg1) is not xpb2.GraphProto:
        _print("Graph pg1 is not an ONNX graph.")
        return False
    if type(pg2) is not xpb2.GraphProto:
        _print("Graph pg2 is not an ONNX graph.")
        return False
    if type(cg) is not xpb2.GraphProto:
        _print("Graph cg is not an ONNX graph.")
        return False

    # Construct the match list
    io_match = pg1_match
    io_match.extend(pg2_match)

    # Do the joint (2x concat)
    g1 = concat(pg1, pg2, rename_nodes=True, complete=False, _verbose=False, **kwargs)
    g = concat(g1, cg, rename_nodes=True, io_match=io_match, complete=complete, _verbose=False, **kwargs)
    if not g:
        _print("Graph merge failed. Please checkout concat for additional options.", "MSG", (not _verbose))

    return g


def split(
        pg: xpb2.GraphProto,
        cg1: xpb2.GraphProto,
        cg2: xpb2.GraphProto,
        cg1_match: [] = None,
        cg2_match: [] = None,
        complete: bool = True,
        _verbose: bool = True,
        **kwargs):
    """
    split takes takes a single parent and matches the outputs to the inputs of two childs (cg1 & cg2)

    Split matches the outputs of pg to the inputs of cg1 and cg2 as specified in cg1_match and cg2_match.
    Desired matches are specified in pairs: [("out1","in1"), ("out2","in2"),...].

    Split by default assumes the resulting joined graph to be complete. Split is merely a wrapper around concat (used
    twice). For more flexible combinations of graphs please see concat().

    Note: ONNX concat operations might give unexpected results if names of elements collide, please use postfix_names()
    to prevent this (and always critically inspect the resulting graph).

    Args:
        pg: The parent graph
        cg1: The left child.
        cg2: The right child.
        cg1_match: (Optional) List of pairs matching outputs of pg to inputs of cg1. Default [].
        cg2_match: (Optional) List of pairs matching outputs of pg to inputs of cg2. Default [].
        complete: (Optional) Boolean indicating whether the resulting graph should be complete (i.e., should pass check). Default True.
        _verbose: (Optional) Boolean indicating whether or not verbose user feedback should be provided. Default True.
    Returns:
        The joined graph g (of False is something fails along the way).
    """
    # immutable defaults:
    if cg1_match is None:
        cg1_match = []
    if cg2_match is None:
        cg2_match = []

    # prevent changes to original
    pg = copy.deepcopy(pg)
    cg1 = copy.deepcopy(cg1)
    cg2 = copy.deepcopy(cg2)

    if type(pg) is not xpb2.GraphProto:
        _print("Graph pg is not an ONNX graph.")
        return False
    if type(cg1) is not xpb2.GraphProto:
        _print("Graph cg1 is not an ONNX graph.")
        return False
    if type(cg2) is not xpb2.GraphProto:
        _print("Graph cg2 is not an ONNX graph.")
        return False

    # Create the split (using concat 2x)
    g1 = concat(pg, cg1, rename_nodes=True, io_match=cg1_match, complete=False, _verbose=False, **kwargs)
    g = concat(g1, cg2, rename_nodes=True, io_match=cg2_match, complete=complete, _verbose=False, **kwargs)
    if not g:
        _print("Graph merge failed. Please checkout concat() for additional options.", "MSG", (not _verbose))

    return g


def concat(
        sg1: xpb2.GraphProto,
        sg2: xpb2.GraphProto,
        complete: bool = False,
        rename_nodes: bool = True,
        io_match: [] = None,
        rename_io: bool = False,
        edge_match: [] = None,
        rename_edges: bool = False,
        rename_init: bool = False,
        _verbose: bool = True,
        **kwargs):
    """
    concat concatenates two graphs.

    Concat is the flexible (but also rather complex) workhorse for the merge, join, and split functions and
    can be used to quite flexibly paste together two (sub)graphs. Contrary to merge, join, and split, concat
    does not by default assume the resulting onnx graph to be complete (i.e., to contain inputs and outputs and to
    pass check()), and it can thus be used as an intermediate function when constructing larger graphs.

    Concat is flexible and versatile, but it takes time to master. See example_merge.py in the examples folder
    for a number of examples.

    Args:
        sg1: Subgraph 1, the parent.
        sg2: Subgraph 2, the child.
        complete: (Optional) Boolean indicating whether the resulting graph should be checked using so.check(). Default False.
        rename_nodes: (Optional) Boolean indicating whether the names of the nodes in the graph should be made unique. Default True.
        io_match: (Optional) Dict containing pairs of outputs of sg1 that should be matched to inputs of sg2. Default [].
        rename_io: (Optional) Boolean indicating whether the inputs and outputs of the graph should be renamed. Default False.
        edge_match: (Optional) Dict containing pairs edge names of sg1 (i.e., node outputs) that should be matched to edges of sg2 (i.e., node inputs). Default [].
        rename_edges: (Optional) Boolean indicating whether the edges should be renamed (default False)
        _verbose: (Optional) Boolean indicating whether verbose output should be printed (default False)
    Returns:
        The concatenated graph g, or False if something goes wrong along the way.
    """
    # immutable defaults:
    if io_match is None:
        io_match = []
    if edge_match is None:
        edge_match = []

    # prevent changes to original
    sg1 = copy.deepcopy(sg1)
    sg2 = copy.deepcopy(sg2)

    # Check input types:
    if type(sg1) is not xpb2.GraphProto:
        _print("Graph sg1 is not an ONNX graph. Abort.")
        return False
    if type(sg2) is not xpb2.GraphProto:
        _print("Graph sg2 is not an ONNX graph. Abort.")
        return False

    # Rename node names if requested (default True)
    if rename_nodes:
        _print("Renaming node names in graph.", "MSG", (not _verbose))
        sg1 = postfix_names(sg1, "_sg1", "node")
        sg2 = postfix_names(sg2, "_sg2", "node")

    if io_match:
        _print("Matching specified inputs and outputs..", "MSG", (not _verbose))
        for io_pair in io_match:
            for outputs in sg1.output:
                if outputs.name == io_pair[0]:
                    sg1 = delete_output(sg1, io_pair[0])
            for inputs in sg2.input:
                if inputs.name == io_pair[1]:
                    sg2 = delete_input(sg2, io_pair[1])
            for item in sg2.node:
                for index, name in enumerate(item.input):
                    if name == io_pair[1]:
                        item.input[index] = io_pair[0]

    if rename_io:
        _print("Renaming inputs and outputs.", "MSG", (not _verbose))
        sg1 = postfix_names(sg1, "_sg1", "io")
        sg2 = postfix_names(sg2, "_sg2", "io")

    if edge_match:
        _print("Matching edges.", "MSG", (not _verbose))
        for edge_pair in edge_match:
            for item in sg2.node:
                for index, name in enumerate(item.input):
                    if name == edge_pair[1]:
                        item.input[index] = edge_pair[0]

    if rename_edges:
        _print("Renaming edges.", "MSG", (not _verbose))
        sg1 = postfix_names(sg1, "_sg1", "edge")
        sg2 = postfix_names(sg2, "_sg2", "edge")

    if rename_init:
        _print("Renaming init.", "MSG", (not _verbose))
        sg1 = postfix_names(sg1, "_sg1", "init")
        sg2 = postfix_names(sg2, "_sg2", "init")

    # Paste graphs together:
    _print("Pasting graphs.", "MSG", (not _verbose))
    g = _paste_graphs(sg1, sg2)

    if complete:
        if not check(g, _verbose=_verbose, **kwargs):
            _print("The end result does not pass check(). Are you sure you want a complete result? Set complete=False "
                   "to continue concat without checking.")
            return False

    return g


def postfix_names(
        g: xpb2.GraphProto,
        postfix: str = "_g1",
        elem: str = "node"):
    """
    postfix_names is a utility function used by concat() to rename parts of an onnx graph.

    When merging (or otherwise manipulating) onnx graphs it is often useful to create unique names of the
    various elements of the graph. This function postfixes each name in supplied graph g of elements of type elem
    by the supplied postfix.

    Args:
        g: The graph
        postfix: (Optional) The postfix for the names of the elements. Default "_g1".
        elem: (Optional) The type of element. Options are "node", "init", "edge", "input", "output", "io", and "all". Default "node".
    """
    if elem == 'node':
        for item in g.node:
            item.name = item.name + postfix
        return g
    elif elem == 'init':
        for init in g.initializer:
            init.name = init.name + postfix
        return g
    elif elem == 'edge':
        for init in g.node:
            for index, name in enumerate(init.input):
                init.input[index] = init.input[index] + postfix
            for index, name in enumerate(init.output):
                init.output[index] = init.output[index] + postfix
        return g
    elif elem == 'input':
        for item in g.input:
            item.name = item.name + postfix
        return g
    elif elem == 'output':
        for item in g.output:
            item.name = item.name + postfix
        return g
    elif elem == 'io':
        cg = postfix_names(g, postfix, "input")
        cg = postfix_names(cg, postfix, "output")
        return cg
    elif elem == 'all':
        cg = postfix_names(g, postfix, "node")
        cg = postfix_names(cg, postfix, "init")
        cg = postfix_names(cg, postfix, "edge")
        cg = postfix_names(cg, postfix, "input")
        cg = postfix_names(cg, postfix, "output")
        return cg
    else:
        _print("No names have been changed; did you select the right element?", "MSG")

    return g


def _paste_graphs(
        sg1: xpb2.GraphProto,
        sg2: xpb2.GraphProto):
    """
    _paste_graphs takes two subgraphs and pastes all of their objects into a single graph.

    Note, _paste_graphs does not conduct any checks, it just blindly copies. It is used internally
    by the concat() function.

    Args:
        sg1: The first subgraph
        sg2: The second subgraph

    Returns:
        g: the joined graph
    """
    g = copy.deepcopy(sg1)

    # Copy initializers from sg2
    for init in sg2.initializer:
        g.initializer.append(init)

    # Copy nodes from sg2
    for node in sg2.node:
        g.node.append(node)

    # Copy inputs and outputs from sg2
    for item in sg2.input:
        g.input.append(item)
    for item in sg2.output:
        g.output.append(item)

    return g



# Executed when running
# > python sclblonnx
# from terminal.
if __name__ == '__main__':
    print("The sclblonnx package currently does not have any command line options.")
    print("Please import the package into your python project and work from there.")



import os
import numpy as np
from onnx import onnx_ml_pb2 as xpb2
from sclblonnx import empty_graph, graph_from_file, graph_to_file, run, list_data_types, list_operators, sclbl_input


def test_empty_graph():
    g = empty_graph()
    assert type(g) is xpb2.GraphProto, "Failed to create empty graph."


def test_graph_from_file():
    g = graph_from_file("files/non-existing-file.onnx")
    assert not g, "Graph from file failed to check emtpy file."
    g = graph_from_file("files/example01.onnx")
    assert type(g) is xpb2.GraphProto, "Graph from file failed to open file."


def test_graph_to_file():
    g = empty_graph()
    check1 = graph_to_file(g, "")
    assert not check1, "Graph to file failed should have failed."
    check2 = graph_to_file(g, "files/test_graph_to_file.onnx")
    assert check2, "Graph to file failed to write file."
    os.remove("files/test_graph_to_file.onnx")


def test_run():
    g = graph_from_file("files/add.onnx")
    example = {"x1": np.array([2]).astype(np.float32), "x2": np.array([5]).astype(np.float32)}
    result = run(g,
                    inputs=example,
                    outputs=["sum"]
                    )
    assert result[0] == 7, "Add output not correct."
    result = run(g, inputs="", outputs="sum")
    assert not result, "Model with this input should not run."


def test_display():
    from onnx import TensorProto
    print(TensorProto.DOUBLE)

    return True  # No test for display


def test_scblbl_input():
    example = {"in": np.array([1,2,3,4]).astype(np.int32)}
    result = sclbl_input(example, _verbose=False)
    assert result == '{"input": "CAQQBkoQAQAAAAIAAAADAAAABAAAAA==", "type":"pb"}', "PB output not correct."

    example = {"x1": np.array([1,2,3,4]).astype(np.int32), "x2": np.array([1,2,3,4]).astype(np.int32)}
    result = sclbl_input(example, _verbose=False)
    assert result == '{"input": ["CAQQBkoQAQAAAAIAAAADAAAABAAAAA==","CAQQBkoQAQAAAAIAAAADAAAABAAAAA=="], "type":"pb"}',\
        "PB output 2 not correct. "

    example = {"in": np.array([1,2,3,4]).astype(np.int32)}
    result = sclbl_input(example, "raw", _verbose=False)
    assert result == '{"input": "AQAAAAIAAAADAAAABAAAAA==", "type":"raw"}', "Raw output not correct."

    example = {"x1": np.array([1,2,3,4]).astype(np.int32), "x2": np.array([1,2,3,4]).astype(np.int32)}
    result = sclbl_input(example, "raw", _verbose=False)
    assert result == '{"input": ["AQAAAAIAAAADAAAABAAAAA==","AQAAAAIAAAADAAAABAAAAA=="], "type":"raw"}',\
        "Raw output 2 not correct. "

    example = {"x1": np.array([1.2]).astype(np.float32), "x2": np.array([2.5]).astype(np.float32)}
    input = sclbl_input(example, _verbose=False)
    print(input)


def test_list_data_types():
    test = list_data_types()
    assert test, "Data types should be listed."


def test_list_operators():
    test = list_operators()
    assert test, "Operators should be listed."


from onnx import helper as xhelp
from onnx import onnx_ml_pb2 as xpb2
from sclblonnx.utils import _value, _data_type, _parse_element, _print


# list_inputs lists all inputs of a graph
def list_inputs(graph: xpb2.GraphProto):
    """ Tries to list the inputs of a given graph.

    Args:
        graph the ONNX graph
    """
    if type(graph) is not xpb2.GraphProto:
        _print("graph is not a valid ONNX graph.")
        return False

    i = 1
    for elem in graph.input:
        name, dtype, shape = _parse_element(elem)
        print("Input {}: Name: '{}', Type: {}, Dimension: {}".format(i, name, dtype, shape))
        i += 1

    if i == 1:
        print("No inputs found.")

    return True


# add_input adds an input to a graph
def add_input(
        graph: xpb2.GraphProto,
        name: str,
        data_type: str,
        dimensions: [],
        **kwargs):
    """ Add an input to a graph

    Args:
        graph: A graph, onnx.onnx_ml_pb2.GraphProto.
        name: String, the name of the input as used to determine the graph topology.
        data_type: String, the data type of the input. Run list_data_types() for an overview.
        dimensions: List[] specifying the dimensions of the input.
        **kwargs

    Returns:
        The extended graph.

    """
    if type(graph) is not xpb2.GraphProto:
        _print("graph is not a valid ONNX graph.")
        return False

    dtype = _data_type(data_type)
    if not dtype:
        return False

    try:
        graph.input.append(xhelp.make_tensor_value_info(name, dtype, dimensions, **kwargs), *kwargs)
    except Exception as e:
        _print("Unable to add the input: " + str(e))
        return False
    return graph


# rename_input renames an input
def rename_input(graph, current_name, new_name):
    """ Rename an input to a graph

    Args:
        graph: A graph, onnx.onnx_ml_pb2.GraphProto.
        current_name: String, the current input name.
        new_name: String, the name desired input name.

    Returns:
        The changed graph.
    """
    if type(graph) is not xpb2.GraphProto:
        _print("graph is not a valid ONNX graph.")
        return False

    found = False
    for input in graph.input:
        if input.name == current_name:
            input.name = new_name
            found = True
    if not found:
        _print("Unable to find the input to rename.")
        return False

    # And rename it in every nodes that takes this as input:
    for node in graph.node:
        for index, name in enumerate(node.input):
            if name == current_name:
                node.input[index] = new_name

    return graph


# replace input replaces and existing input
def replace_input(
        graph: xpb2.GraphProto,
        name: str,
        data_type: str,
        dimensions: [],
        **kwargs):
    """ Changes an existing input in a graph

    Args:
        graph: A graph, onnx.onnx_ml_pb2.GraphProto.
        name: String, the name of the input as used to determine the graph topology.
        data_type: String, the data type of the input. Run list_data_types() for an overview.
        dimensions: List[] specifying the dimensions of the input.,
        **kwargs

    Returns:
        The extended graph.

    """
    if type(graph) is not xpb2.GraphProto:
        _print("graph is not a valid ONNX graph.")
        return graph

    # Remove the named input
    found = False
    try:
        for elem in graph.input:
            if elem.name == name:
                graph.input.remove(elem)
                found = True
    except Exception as e:
        _print("Unable to iterate the inputs. " + str(e))
        return False
    if not found:
        _print("Unable to find the input by name.")

    # Create the new value
    try:
        val = _value(name, data_type, dimensions, **kwargs)
    except Exception as e:
        _print("Unable to create value. " + str(e))
        return False

    # Add the value to the input
    try:
        graph.input.append(val, *kwargs)
    except Exception as e:
        _print("Unable to add the input: " + str(e))
        return False

    return graph


# delete_input deletes an existing input
def delete_input(
        graph: xpb2.GraphProto,
        name: str):
    """ Removes an existing input of a graph by name

    Args:
        graph: A graph, onnx.onnx_ml_pb2.GraphProto.
        name: String, the name of the input as used to determine the graph topology.

    Returns:
        The extended graph.

    """
    if type(graph) is not xpb2.GraphProto:
        return graph

    # Remove the named output
    found = False
    try:
        for elem in graph.input:
            if elem.name == name:
                graph.input.remove(elem)
                found = True
    except Exception as e:
        _print("Unable to iterate the inputs. " + str(e))
        return False
    if not found:
        _print("Unable to find the input by name.")
        return False

    return graph



"""
The sclblonnx package provides onnx tools
"""
# Ran on import of the package. Check version:
import sys

if sys.version_info < (3, 0):
    print('Sclblonnx requires Python 3, while Python ' + str(sys.version[0] + ' was detected. Terminating... '))
    sys.exit(1)

from .version import __version__

from .main import \
    empty_graph, \
    graph_from_file, \
    graph_to_file, \
    run, \
    display, \
    sclbl_input, \
    list_data_types, \
    list_operators

from .validate import \
    clean, \
    check

from .node import \
    node, \
    add_node, \
    add_nodes, \
    delete_node

from .constant import \
    constant, \
    add_constant

from .input import \
    list_inputs, \
    add_input, \
    rename_input, \
    replace_input, \
    delete_input

from .output import \
    list_outputs, \
    add_output, \
    rename_output, \
    replace_output, \
    delete_output

from .merge import \
    merge, \
    join, \
    split, \
    concat, \
    postfix_names










import base64
import json
import os
import subprocess
import onnxruntime as xrt
from onnx import ModelProto as xmp
from onnx import helper as xhelp
from onnx import onnx_ml_pb2 as xpb2
from onnx import save as xsave
from onnx import numpy_helper as xnp
import onnx
import sclblonnx._globals as glob
from sclblonnx.utils import _print


# empty_graph creates an empty graph
def empty_graph(
        _default_name: str = "sclblgraph"):
    """ empty_graph returns an empty graph

    Note, an empty graph does not pass the check() as it does not contain input and output.

    Args:
        _default_name: Graph name, default sclblgraph

    Returns:
        An empty graph
    """
    try:
        graph = xpb2.GraphProto(name=_default_name)
    except Exception as e:
        _print("Unable to create graph: " + str(e))
        return False
    return graph


# graph_from_file opens an existing onnx file and returns the graph
def graph_from_file(
        filename: str):
    """ Retrieve a graph object from an onnx file

    Function attempts to open a .onnx file and returns its graph.

    Args:
        filename: String indicating the filename / relative location.

    Returns:
        An ONNX graph or False if unable to open.

    """
    mod_temp = xmp()
    try:
        with open(filename, 'rb') as fid:
            content = fid.read()
            mod_temp.ParseFromString(content)
        graph = mod_temp.graph
    except Exception as e:
        _print("Unable to open the file: " + str(e))
        return False
    return graph


# graph_to_file saves a graph to a file
def graph_to_file(
        graph: xpb2.GraphProto,
        filename: str,
        _producer: str = "sclblonnx",
        onnx_opset_version = 12,
        **kwargs):
    """ graph_to_file stores an onnx graph to a .onnx file

    Stores a graph to a file

    Args:
        graph: An onnx graph
        filename: The filename of the resulting file
        _producer: Optional string with producer name. Default 'sclblonnx'
        onnx_opset_version: Optional version number for ONNX opset. Default 12
    Returns:
        True if successful, False otherwise.
    """
    if not filename:
        _print("Unable to save: Please specify a filename.")
        return False

    if type(graph) is not xpb2.GraphProto:
        _print("Unable to save: Graph is not an ONNX graph")

    try:
        if not 'opset_imports' in kwargs:
            op = onnx.OperatorSetIdProto()
            op.version = onnx_opset_version
            mod = xhelp.make_model(graph, producer_name=_producer, opset_imports=[op], **kwargs)
        else:
            mod = xhelp.make_model(graph, producer_name=_producer, **kwargs)
    except Exception as e:
        print("Unable to convert graph to model: " + str(e))
        return False

    try:
        xsave(mod, filename, **kwargs)
    except Exception as e:
        print("Unable to save the model: " + str(e))
        return False

    return True


# run executes a given graph and returns its result
def run(
        graph: xpb2.GraphProto,
        inputs: {},
        outputs: [],
        _tmpfile: str = ".tmp.onnx",
        onnx_opset_version = 12,
        **kwargs):
    """ run executes a give graph with the given input and returns the output

    Args:
        graph: The onnx graph
        inputs: an object with the named inputs; please check the data types
        outputs: list of named outputs
        _tmpfile: String the temporary filename for the onnx file to run.
        onnx_opset_version: Optional version number for ONNX opset. Default 12
        
    Returns:
        The result (or False if it fails somewhere)
        """

    store = graph_to_file(graph, _tmpfile, onnx_opset_version=onnx_opset_version)
    if not store:
        _print("Unable to store model for evaluation.")
        return False

    try:
        sess = xrt.InferenceSession(_tmpfile, **kwargs)
        out = sess.run(outputs, inputs)
    except Exception as e:
        _print("Failed to run the model: " + str(e))
        return False

    try:
        os.remove(_tmpfile)
    except Exception as e:
        print("We were unable to delete the file " + _tmpfile, "MSG")

    return out


# display uses Netron to display a graph
def display(
        graph: xpb2.GraphProto,
        _tmpfile: str = '.tmp.onnx'):
    """ display a onnx graph using netron.

    Pass a graph to the display function to open it in Netron.
    Note: Due to the complexities of cross platform opening of source and the potential lack of
    a Netron installation this function might not always behave properly.
    Note2: This function might leave a file called .temp.onnx if it fails to remove the file.

    Args:
        graph: an ONNX graph
        _tmpfile: an optional string with the temporary file name. Default .tmp.onnx

    Returns:
        True if one of the 3 methods to open the file did not raise any warnings.

    Raises:
        SclblONNXError
    """
    if type(graph) is not xpb2.GraphProto:
        _print("graph is not a valid ONNX graph.")
        return False

    # store as tmpfile
    graph_to_file(graph, _tmpfile)

    file_open = False
    # try open on unix:
    if not file_open:
        try:
            subprocess.run(['xdg-open', _tmpfile])
            file_open = True
        except Exception:
            file_open = False

    # try open on mac:
    if not file_open:
        try:
            subprocess.run(['open', _tmpfile])
            file_open = True
        except Exception:
            file_open = False

    # try open on windows:
    if not file_open:
        try:
            os.startfile(_tmpfile)
            file_open = True
        except Exception:
            file_open = False

    # Result:
    return file_open


# sclbl_input generates the example input for a Scailable runtime
def sclbl_input(
        inputs: {},
        example_type: str = "pb",
        _verbose: bool = True):
    """ input_str returns an example input for a Scailable runtime

    The method takes a valid input object to an onnx graph (i.e., one used for the "inputs" argument
    in the run() function, and returns and prints an example input to a Scailable runtime / REST endpoint

    Args:
        inputs: The input object as supplied to the run() function to test an ONNX grph
        example_type: The type of example string ("raw" for base64 encoded, or "pb" for protobuf, default pb)
        _verbose: Print user feedback; default True (note, errors are always printed).

    Returns:
        An example input to a Scailable runtime.
    """
    if not inputs:
        _print("No input provided.")

    if example_type == "raw":
        if len(inputs) == 1:
            for val in inputs.values():
                bytes = val.tobytes()
                encoded = base64.b64encode(bytes)
                value_str = '"'+encoded.decode('ascii')+'"'
        else:
            value_str = '["'
            for val in inputs.values():
                bytes = val.tobytes()
                encoded = base64.b64encode(bytes)
                value_str += (encoded.decode('ascii') + '","')
            value_str = value_str.rstrip(',"')
            value_str += '"]'

        input_json = '{"input": ' + value_str + ', "type":"raw"}'
        if _verbose:
            _print("The following input string can be used for the Scailable runtime:", "MSG")
            _print(input_json, "LIT")
        return input_json

    elif example_type == "pb" or "protobuf":

        if len(inputs) == 1:
            for val in inputs.values():
                tensor = xnp.from_array(val)
                serialized = tensor.SerializeToString()
                encoded = base64.b64encode(serialized)
                value_str = '"'+encoded.decode('ascii')+'"'
        else:
            value_str = '["'
            for val in inputs.values():
                tensor = xnp.from_array(val)
                serialized = tensor.SerializeToString()
                encoded = base64.b64encode(serialized)
                value_str += (encoded.decode('ascii') + '","')
            value_str = value_str.rstrip(',"')
            value_str += '"]'

        input_json = '{"input": ' + value_str + ', "type":"pb"}'
        if _verbose:
            _print("The following input string can be used for the Scailable runtime:", "MSG")
            _print(input_json, "LIT")
            _print("The following input string can be used for the web front-end:", "MSG")
            _print(value_str, "LIT")
        return input_json


# list_data_types prints all available data types
def list_data_types():
    """ List all available data types. """
    _print(json.dumps(glob.DATA_TYPES, indent=2), "MSG")
    _print("Note: STRINGS are not supported at this time.", "LIT")
    return True


# list_operators prints all operators available within Scailable
def list_operators():
    """ List all available Scailable ONNX operators. """
    try:
        with open(glob.VERSION_INFO_LOCATION, "r") as f:
            glob.ONNX_VERSION_INFO = json.load(f)
    except FileNotFoundError:
        print("Unable to locate the ONNX_VERSION INFO.")
        return False
    _print(json.dumps(glob.ONNX_VERSION_INFO['operators'], indent=2), "MSG")
    return  True


# No command line options for this script:
if __name__ == '__main__':
    print("No command line options available for main.py.")



import onnxoptimizer
from onnx import __version__ as xversion
from onnx import checker
import onnx
from onnx import helper as xhelp
from onnx import onnx_ml_pb2 as xpb2
from onnxsim import simplify
from packaging import version

import sclblonnx._globals as glob
from sclblonnx.utils import _load_version_info, _print


# clean cleans a graph if possible (but also provides a stringent check)
def clean(
        graph: xpb2.GraphProto,
        _optimize: bool = True,
        _simplify: bool = True,
        _remove_initializer: bool = True,
        _producer: str = "sclblonnx",
        _verbose: bool = True,
        **kwargs):
    """ clean cleans an ONNX graph using onnx tooling

    This method will attempt to clean the supplied graph by
    a. Removing initializers from input
    b. Optimizing it using onnxoptimizer.optimize
    c. Simplifying it using onnxsim.simplify

    If one of these fails the method will print an error message and return the unaltered graph.

    Args:
        graph: An ONNX graph
        _optimize: Boolean, default True. Optimize the model using onnxoptimizer.
        _simplify: Boolean, default True. Simplify the model using simplify.
        _remove_initializer: Boolean, default True. Remove initializers from input.
        _producer: Optional string with producer name. Default 'sclblonnx' (used for internal conversion)
        _verbose: Print user feedback; default True (note, errors are always printed).
        **kwargs

    Returns:
        The cleaned ONNX graph, or the old graph if an error occurs.
    """
    try:
        if not 'opset_imports' in kwargs:
            op = onnx.OperatorSetIdProto()
            op.version = 12
            mod = xhelp.make_model(graph, producer_name=_producer, opset_imports=[op], **kwargs)
        else:
            mod = xhelp.make_model(graph, producer_name=_producer, **kwargs)
    except Exception as e:
        _print("Unable to create the model: " + str(e))
        return graph

    if _optimize:
        try:
            mod = onnxoptimizer.optimize(mod, glob.OPTIMIZER_PASSES, **kwargs)
        except Exception as e:
            _print("Unable to optimize your model: " + str(e))
            return graph

    if _simplify:
        try:
            mod, _ = simplify(mod, **kwargs)
        except Exception as e:
            _print("Unable to simplify your model: " + str(e))
            return graph

    # From: onnxruntime/tools/python/remove_initializer_from_input.py
    graph = mod.graph
    if _remove_initializer:
        inputs = graph.input
        name_to_input = {}
        for input in inputs:
            name_to_input[input.name] = input
        for initializer in graph.initializer:
            if initializer.name in name_to_input:
                inputs.remove(name_to_input[initializer.name])

    _print("The graph was successfully cleaned.", "MSG", (not _verbose))
    return graph


# check checks the graph and inspects whether it is valid.
def check(
        graph: xpb2.GraphProto,
        _producer: str = "sclblonnx",
        _onnx_check: bool = True,
        _sclbl_check: bool = True,
        _verbose: bool = True,
        **kwargs):
    """ check whether or not an existing graph can be converted using the Scailable platform

    We assume that a user will use graph_to_file() in this package to store the model. This

     Args:
        graph: an ONNX graph
        _producer: String optional
        _onnx_check: Bool, default True. Run ONNX checker.check().
        _sclbl_check: Bool, default True.  Run Scailable checks.
        _verbose: Print user feedback; default True (note, errors are always printed).
        **kwargs

    Returns:
        True if the graph passes all the test. False otherwise.
    """
    # Check if this is a valid graph:
    if type(graph) is not xpb2.GraphProto:
        _print("Graph is not a valid ONNX graph.")
        return False

    # Convert to model:
    try:
        if not 'opset_imports' in kwargs:
            op = onnx.OperatorSetIdProto()
            op.version = 12
            mod = xhelp.make_model(graph, producer_name=_producer, opset_imports=[op], **kwargs)
        else:
            mod = xhelp.make_model(graph, producer_name=_producer, **kwargs)
    except Exception as e:
        _print("Unable to create the model: " + str(e))
        return False

    # Standard ONNX checking:
    if _onnx_check and False:
        try:
            checker.check_model(mod, **kwargs)
        except Exception as e:
            _print("Model fails on standard ONNX checker: " + str(e))
            return False

    if _sclbl_check:

        # User feedback
        _print("Running Scailable specific checks for WASM conversion. \nUse _sclbl_check=False to turn off", "MSG", (not _verbose))

        # input / output checking:
        if not graph.input:
            _print("This graph does not contain any inputs.")
            return False

        if not graph.output:
            _print("This graph does not contain any outputs.")
            return False

        # Sclbl checking:
        if not glob.ONNX_VERSION_INFO:
            if not _load_version_info():
                _print("Unable to load the ONNX_VERSION INFO.")

        # Check general ONNX version:
        if version.parse(xversion) < version.parse(glob.ONNX_VERSION_INFO['onnx_version']['version_min']):
            _print("Your current onnx version is lower then our support minimum. Please update your ONNX to {}".format(
                glob.ONNX_VERSION_INFO['onnx_version']['version_min']))
            return False

        if version.parse(xversion) > version.parse(glob.ONNX_VERSION_INFO['onnx_version']['version_max']):
            _print(
                "Your current onnx version is higher then our support max. Please downgrade your ONNX version to {}".format(
                    glob.ONNX_VERSION_INFO['onnx_version']['version_max']))
            return False

        if mod.ir_version < glob.ONNX_VERSION_INFO['onnx_version']['ir_version_min']:
            _print("Your current IR version is lower then our support minimum. Please update to {}".format(
                glob.ONNX_VERSION_INFO['onnx_version']['ir_version_min']))
            return False

        if mod.ir_version > glob.ONNX_VERSION_INFO['onnx_version']['ir_version_max']:
            _print(
                "Your current IR version is higher then our support max. Please downgrade to {}".format(
                    glob.ONNX_VERSION_INFO['onnx_version']['ir_version_max']))
            return False

        # Interate through opset and check:
        for key in mod.opset_import:
            v = key.version
            if v < glob.ONNX_VERSION_INFO['onnx_version']['opset_min']:
                _print("One or more operators use an opset version that is too low. Please update to {}".format(
                    glob.ONNX_VERSION_INFO['onnx_version']['opset_min']))
                return False

            if v > glob.ONNX_VERSION_INFO['onnx_version']['opset_max']:
                _print(
                    "One or more operators use an opset version that is too high. Please downgrade to {}".format(
                        glob.ONNX_VERSION_INFO['onnx_version']['opset_max']))
                return False

        # Check individual nodes:
        not_supported = []
        for n in graph.node:
            op = n.op_type
            if op not in glob.ONNX_VERSION_INFO['operators']:
                not_supported.append(op)
        if not_supported:
            _print("The operator(s) {} are currently not supported.".format(not_supported))
            return False

        # Check dynamic
        for inputs in graph.input:
            if not inputs.type.tensor_type.shape.dim:
                _print("Your graph contains dynamically sized inputs, this is currently not supported.")
                return False
            for elem in inputs.type.tensor_type.shape.dim:
                if elem.dim_value == 0 or elem.dim_value == "":
                    _print("Your graph contains dynamically size inputs, this is currently not supported.")

    if not _sclbl_check and not _onnx_check:
        _print("Set _sclbl_check or _onnx_check to True to run any checks.")

    _print("Your graph was successfully checked.", "MSG", (not _verbose))
    return True



import json
from onnx import helper as xhelp
from onnx import onnx_ml_pb2 as xpb2
import sclblonnx._globals as glob


# _parse_element parses a graph input  or output element and return its properties for printing.
def _parse_element(elem: xpb2.ValueInfoProto):
    """ Parse a graph input or output element and return its contents.

    Utility.

    Args:
        elem, a ValueInfoProto.

    Returns:
        name The name of the element
        data_type The data type of the element
        shape_str The dimensions of the element
    """
    name = getattr(elem, 'name', "None")
    data_type = "NA"
    shape_str = "NA"
    etype = getattr(elem, 'type', False)
    if etype:
        ttype = getattr(etype, 'tensor_type', False)
        if ttype:
            data_type = _data_string(getattr(ttype, 'elem_type', 0))
            shape = getattr(elem.type.tensor_type, "shape", False)
            if shape:
                shape_str = "["
                dims = getattr(shape, 'dim', [])
                for dim in dims:
                    vals = getattr(dim, 'dim_value', "?")
                    shape_str += (str(vals) + ",")
                shape_str = shape_str.rstrip(",")
                shape_str += "]"
    return name, data_type, shape_str


# _value creates a new value description
def _value(name: str,
           data_type: str,
           dimensions: [],
           **kwargs):
    """_value creates a tensor value

    Args:
        name: TensorValue name.
        data_type: String data type.
        dimensions: List with dimensions.
        **kwargs


    """
    dtype = _data_type(data_type)
    if not dtype:  # Error printed by _data_type()
        return False

    try:
        val = xhelp.make_tensor_value_info(name, dtype, dimensions, **kwargs)
    except Exception as e:
        _print("Unable to create tensor value: " + str(e))
        return False

    return val


def _input_details(graph: xpb2.GraphProto):
    """ List details of the inputs of the graph by name.

    Args:
        The graph object.

    Returns:
        A dict containing the input details.

    """
    if type(graph) is not xpb2.GraphProto:
        _print("graph is not a valid ONNX graph.")
        return False

    names = {}
    for elem in graph.input:
        name, dtype, shape = _parse_element(elem)
        desc = {"data_type": dtype, "shape": shape}
        names[name] = desc

    return names


def _output_details(graph: xpb2.GraphProto):
    """ List details of the outputs of the graph by name.

    Args:
        The graph object.

    Returns:
        A dict containing the output details.

    """
    if type(graph) is not xpb2.GraphProto:
        _print("graph is not a valid ONNX graph.")
        return False

    names = {}
    for elem in graph.output:
        name, dtype, shape = _parse_element(elem)
        desc = {"data_type": dtype, "shape": shape}
        names[name] = desc

    return names


# bcolors, used for printing.
class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


# _print controls the printing of user feedback throughout the package
def _print(
        msg: str,
        print_type: str = "ERR",
        silent: bool = False):
    """ print controls the printing throughout the package.

    Args:
        msg: The message to print
        print_type: "MSG", "LIT" or "ERR", default "ERR"
        silent: Suppress printing, default False
    """
    if not silent:
        if print_type == "MSG":
            print(msg)
        elif print_type == "ERR":
            print(f"{bcolors.FAIL}ERROR: "+msg+f"{bcolors.ENDC}")
        elif print_type == "LIT":
            print(f"{bcolors.OKGREEN}" + msg + f"{bcolors.ENDC}")
        else:
            print(msg)


# _load_version_info loads info of supported ONNX versions (see _globals.py)
def _load_version_info() -> bool:
    """Loads the version info json.
    Function opens and parses the file supported_onnx.json in the current
    package folder to check current Scailable toolchain requirements.

    Note: the supported models are loaded into the glob.ONNX_VERSION_INFO
    dictionary to make them available to the whole package.

    Args:

    Returns:
        True if the supported version info is successfully loaded.
    """
    try:
        with open(glob.VERSION_INFO_LOCATION, "r") as f:
            glob.ONNX_VERSION_INFO = json.load(f)
    except FileNotFoundError:
        _print("Unable to locate the ONNX_VERSION INFO.")
        return False
    return True


# _data_type converts a data_string to the data_type int (see _globals.py)
def _data_type(data_string: str):
    """ convert the data type string (i.e., FLOAT, INT16, etc.) to the appropriate int.

    See: https://deeplearning4j.org/api/latest/onnx/Onnx.TensorProto.DataType.html
    """
    for key, val in glob.DATA_TYPES.items():
        if key == data_string:
            return val
    _print("Data string not found. Use `list_data_types()` to list all supported data strings.")
    return False


# _data_string converts a data_type int to a data string
def _data_string(data_type: int):
    """ convert the data type number to the appropriate string

    See: https://deeplearning4j.org/api/latest/onnx/Onnx.TensorProto.DataType.html
    """
    for key, val in glob.DATA_TYPES.items():
        if val == data_type:
            return key

    _print("Data type not found. Use `list_data_types()` to list all supported data types.")
    return False



import setuptools

# get long description
with open("README.md", "r") as fh:
    long_description = fh.read()

exec(open('sclblonnx/version.py').read())

setuptools.setup(
    name="sclblonnx",
    version=__version__,
    author="Maurits Kaptein",
    author_email="maurits.kaptein@scailable.net",
    description="Python package containing Scailable ONNX tools.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    include_package_data=True,
    url="https://github.com/scailable/sclblonnx/",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'numpy',
        'onnxruntime',
        'onnx>=1.7.0,<=1.12.0',
        'requests',
        'onnxoptimizer',
        'onnx-simplifier',
        'packaging'
      ],
    python_requires='>=3.7',
)



from onnx import helper as xhelp
from onnx import onnx_ml_pb2 as xpb2
import sclblonnx._globals as glob
from sclblonnx.utils import _print


# Node creates a new node
def node(
        op_type: str,
        inputs: [],
        outputs: [],
        name: str = "",
        **kwargs):
    """ Create a new node

    Args:
        op_type: Operator type, see https://github.com/onnx/onnx/blob/master/docs/Operators.md
        inputs: [] list of inputs (names to determine the graph topology)
        outputs: [] list of outputs (names to determine the graph topology)
        name: The name of this node (Optional)
        **kwargs
    """
    if not name:
        name = "sclbl-onnx-node" + str(glob.NODE_COUNT)
        glob.NODE_COUNT += 1

    try:
        node = xhelp.make_node(op_type, inputs, outputs, name, **kwargs)
    except Exception as e:
        _print("Unable to create node: " + str(e))
        return False
    return node


# add_node adds a node to a graph
def add_node(
        graph: xpb2.GraphProto,
        node: xpb2.NodeProto,
        **kwargs):
    """ Add node appends a node to graph g and returns the extended graph

    Prints a message and returns False if fails.

    Args:
        graph: A graph, onnx.onnx_ml_pb2.GraphProto.
        node: A node, onnx.onnx_ml_pb2.NodeProto.
        **kwargs

    Returns:
        The extended graph.
    """
    if type(graph) is not xpb2.GraphProto:
        _print("The graph is not a valid ONNX graph.")
        return False

    if type(node) is not xpb2.NodeProto:
        _print("The node is not a valid ONNX node.")
        return False

    try:
        graph.node.append(node, **kwargs)
    except Exception as e:
        _print("Unable to extend graph: " + str(e))
        return False
    return graph


def add_nodes(
        graph: xpb2.GraphProto,
        nodes: [xpb2.NodeProto],
        **kwargs):
    """ Add a list of nodes appends a node to graph g and returns the extended graph

    Prints a message and returns False if fails.

    Args:
        graph: A graph, onnx.onnx_ml_pb2.GraphProto.
        nodes: A list of nodes, [onnx.onnx_ml_pb2.NodeProto].
        **kwargs

    Returns:
        The extended graph.
    """
    if type(graph) is not xpb2.GraphProto:
        print("graph is not a valid ONNX graph.")
        return False

    for node in nodes:  # error handling in add_node
        graph = add_node(graph, node, **kwargs)
        if not graph:
            return False

    return graph


# delete_node deletes a node from a graph
def delete_node(
        graph: xpb2.GraphProto,
        node_name: str = "",
        **kwargs):
    """ Add node appends a node to graph g and returns the extended graph

    Prints a message and returns False if fails.

    Args:
        graph: A graph, onnx.onnx_ml_pb2.GraphProto.
        node_name: Name of the node to remove.
        **kwargs

    Returns:
        The extended graph.
    """
    if type(graph) is not xpb2.GraphProto:
        _print("The graph is not a valid ONNX graph.")
        return False

    if not node_name:
        _print("Please specify a node name.")
        return False

    found = False
    try:
        for elem in graph.node:
            if elem.name == node_name:
                graph.node.remove(elem)
                found = True
    except Exception as e:
        _print("Unable to iterate the nodes. " + str(e))
        return False
    if not found:
        _print("Unable to find the node by name.")
        return False

    return graph
